//! Token values and data structs.

use std::fmt::Display;

/// Exhaustive enumeration of all types of different tokens.
#[derive(Copy, Clone, Debug, PartialEq, Eq)]
pub enum TokenType {
    // Single-character tokens.
    LeftParen,
    RightParen,
    LeftBrace,
    RightBrace,
    Comma,
    Dot,
    Minus,
    Plus,
    Semicolon,
    Slash,
    Star,

    // One or two character tokens.
    Bang,
    BangEqual,
    Equal,
    EqualEqual,
    Greater,
    GreaterEqual,
    Less,
    LessEqual,

    // Literals.
    Identifier,
    String,
    Number,

    // Keywords.
    And,
    Class,
    Else,
    False,
    Fun,
    For,
    If,
    Nil,
    Or,
    Print,
    Return,
    Super,
    This,
    True,
    Var,
    While,

    Eof,
}

/// Literal value.
#[derive(Clone, Debug, PartialEq)]
pub enum Literal {
    /// String literal.
    String(String),
    // Number literal, represented as f64 (thus it can be decimal).
    Number(f64),
    /// Boolean literal.
    Boolean(bool),
    /// Null literal.
    Nil,
}

/// Consumed token.
#[derive(Clone, Debug, PartialEq)]
pub struct Token {
    /// Type of the token.
    pub token_type: TokenType,
    /// Lexeme that was consumed to create this token.
    pub lexeme: String,
    /// Literal value of the token, if any.
    pub literal: Option<Literal>,
    /// Starting line on which the token was oonsumed from the source.
    pub line: usize,
}

impl Display for Token {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}: {:?} {} {:?}",
            self.line, self.token_type, self.lexeme, self.literal
        )
    }
}
